"""
Module responsible for handling pytoil's interface with the
GitHub REST v3 API.

Author: Tom Fleet
Created: 19/06/2021
"""


from typing import Any, Dict, List, Optional, Set, Union

import httpx

Response = Dict[str, Any]


class API:
    def __init__(self, username: str, token: str) -> None:
        self.username = username
        self.token = token
        self.baseurl = "https://api.github.com/"
        self.headers: Dict[str, str] = {
            "Accept": "application/vnd.github.v3+json",
            "Authorization": f"token {self.token}",
        }

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(username={self.username!r}, token={self.token!r})"
        )

    def get(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> Response:
        """
        Generic authenticated GET request method, used to make a GET
        to any valid GitHub REST v3 API endpoint e.g. 'user/repos'.

        Args:
            endpoint (str): The endpoint to GET.
            params (Optional[Dict[str, Any]], optional): Dictionary of query params to
            pass with the request. Defaults to None.

        Returns:
            Response: JSON Response dict.
        """
        r = httpx.get(url=self.baseurl + endpoint, headers=self.headers, params=params)
        r.raise_for_status()

        response: Response = r.json()

        return response

    def post(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> Response:
        """
        Generic authenticated POST request method, used to make a POST
        to any valid GitHub REST v3 API endpoint.

        Args:
            endpoint (str): The endpoint to POST.
            params (Optional[Dict[str, Any]], optional): Dictionary of query params to
            pass with the request. Defaults to None.

        Returns:
            Response: JSON Response dict.
        """

        r = httpx.post(url=self.baseurl + endpoint, headers=self.headers, params=params)
        r.raise_for_status()

        response: Response = r.json()

        return response

    def create_fork(self, owner: str, repo: str) -> Response:
        """
        Create a fork of the specified repository for the authenticated
        user.

        Args:
            owner (str): Owner of the repo to be forked.
            repo (str): Name of the repo to be forked.

        Returns:
            Response: JSON Response dict.
        """

        return self.post(endpoint=f"repos/{owner}/{repo}/forks")

    def get_repo(self, repo: str) -> Response:
        """
        Get a user's repo by name.

        Args:
            repo (str): Name of the repo to get
                (must be owned by the user or will raise a 404)

        Returns:
            Response: JSON Response dict.
        """

        return self.get(endpoint=f"repos/{self.username}/{repo}")

    def get_repos(self) -> Response:
        """
        Gets all repos owned by the authenticated user.

        Returns:
            Response: JSON Response dict.
        """

        return self.get(endpoint="user/repos", params={"type": "owner"})

    def get_repo_names(self) -> Set[str]:
        """
        Gets the names of all the repos owned by the authenticated user.

        Returns:
            Set[str]: Names of all authenticated user repos.
        """

        # For whatever reason, mypy doesn't like this even though it's totally valid
        return {repo["name"] for repo in self.get_repos()}  # type: ignore

    def get_repo_info(self, repo: str) -> Dict[str, Union[str, int]]:
        """
        Returns a dictionary of key information about a particular repo.

        Info Keys:

        - name
        - description
        - created_at
        - updated_at
        - size
        - license

        Args:
            repo (str): Name of the repo to get info for.

        Returns:
            Dict[str, Union[str, int]]: Dict of repo information.
        """

        raw_repo_data = self.get_repo(repo=repo)

        keys_to_get: List[str] = [
            "name",
            "description",
            "created_at",
            "updated_at",
            "size",
            "license",
        ]

        display_dict: Dict[str, Union[str, int]] = {
            key: raw_repo_data.get(key, "Not found") for key in keys_to_get
        }

        # License is itself a dict
        # Couldn't be bothered doing some clever recursive thing for one key
        if raw_repo_data["license"]:
            display_dict["license"] = raw_repo_data.get("license", "Not Found").get(
                "name", "Not Found"
            )

        return display_dict



"""
Module responsible for handling pytoil's programmatic
interaction with its config file.

Author: Tom Fleet
Created: 18/06/2021
"""

from __future__ import annotations

from dataclasses import field
from pathlib import Path
from typing import List, TypedDict

import yaml
from pydantic.dataclasses import dataclass

from pytoil.config import defaults


class ConfigDict(TypedDict):
    """
    TypedDict for config.

    Exactly the same as the dataclass except projects_dir
    is a str here because that's how it will be brought in
    when deserialising the yaml file.
    """

    projects_dir: str
    token: str
    username: str
    vscode: bool
    common_packages: List[str]
    init_on_new: bool


@dataclass
class Config:
    projects_dir: Path = defaults.PROJECTS_DIR
    token: str = defaults.TOKEN
    username: str = defaults.USERNAME
    vscode: bool = defaults.VSCODE
    common_packages: List[str] = field(default_factory=list)
    init_on_new: bool = defaults.INIT_ON_NEW

    @classmethod
    def from_file(cls, path: Path = defaults.CONFIG_FILE) -> Config:
        """
        Reads in the .pytoil.yml config file and returns
        a populated `Config` object.

        Args:
            path (Path, optional): Path to the config file.
                Defaults to defaults.CONFIG_FILE.

        Returns:
            Config: Populated `Config` object.

        Raises:
            FileNotFoundError: If config file not found.
        """

        try:
            with open(path, mode="r", encoding="utf-8") as f:
                config_dict: ConfigDict = yaml.full_load(f)
        except FileNotFoundError:
            raise
        else:
            return Config.from_dict(config_dict)

    @classmethod
    def from_dict(cls, config_dict: ConfigDict) -> Config:
        """
        Takes in a `ConfigDict` and returns a populated
        `Config` object.

        Args:
            config_dict (ConfigDict): Populated `ConfigDict` object.

        Returns:
            Config: Returned `Config`.
        """

        # We can do this here because pydantic will handle
        # all the conversion between types under the hood
        return Config(**config_dict)  # type: ignore

    @classmethod
    def helper(cls) -> Config:
        """
        Returns a friendly placeholder object designed to be
        written to a config file as a guide to the user on what
        to fill in.

        Most of the fields will be the default but some will have
        helpful instructions.

        Returns:
            Config: Helper config object.
        """

        # Typed ignored here because we know we're correct, mypy doesn't like this
        # probably something to do with Config being a pydantic model
        # not a pure python class
        return Config(
            token="Put your GitHub personal access token here",
            username="This your GitHub username",
        )  # type: ignore

    def to_dict(self) -> ConfigDict:
        """
        Writes out the attributes from the calling instance
        to a dictionary.
        """
        return {
            "projects_dir": str(self.projects_dir),
            "token": self.token,
            "username": self.username,
            "vscode": self.vscode,
            "common_packages": self.common_packages,
            "init_on_new": self.init_on_new,
        }

    def write(self, path: Path = defaults.CONFIG_FILE) -> None:
        """
        Overwrites the config file at `path` with the attributes from
        the calling instance.

        Args:
            path (Path, optional): Config file to overwrite.
                Defaults to defaults.CONFIG_FILE.
        """

        with open(path, mode="w", encoding="utf-8") as f:
            yaml.dump(self.to_dict(), f)

    def can_use_api(self) -> bool:
        """
        Helper method to easily determine whether or not
        the config instance has the required elements
        to use the GitHub API.

        Returns:
            bool: True if can use API, else False.
        """

        conditions = [
            self.username == "",
            self.username == "This your GitHub username",
            self.token == "",
            self.token == "Put your GitHub personal access token here",
        ]

        return not any(conditions)



"""
Module responsible for interacting with local/remote projects.

Author: Tom Fleet
Created: 19/06/2021
"""

from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional

import httpx
import toml

from pytoil.api import API
from pytoil.environments import Conda, Environment, Venv
from pytoil.exceptions import RepoNotFoundError


class Repo:
    def __init__(self, owner: str, name: str, local_path: Path) -> None:
        """
        Representation of a local/remote project.

        Args:
            owner (str): The owner (i.e. GitHub username)
            name (str): Project name
            local_path (Path): The path this repo would have on the local
                filesystem, whether it currently exists or not.
        """
        self.owner = owner
        self.name = name
        self.local_path = local_path

        self.clone_url = f"https://github.com/{self.owner}/{self.name}.git"
        self.html_url = f"https://github.com/{self.owner}/{self.name}"

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(owner={self.owner!r}, "
            + f"name={self.name!r}, "
            + f"local_path={self.local_path!r})"
        )

    def exists_local(self) -> bool:
        """
        Determines whether or not a Repo exists
        on the local filesystem

        Returns:
            bool: True if exists locally, else False.
        """
        return self.local_path.exists()

    def exists_remote(self, api: API) -> bool:
        """
        Determines whether or not a Repo exists
        on GitHub.

        It does this simply by making a GET request
        for the Repo in question and looking at the response
        code.

        Args:
            api (API): pytoil API object.

        Returns:
            bool: True if exists remote, else False.
        """
        try:
            # Capture any response and throw it away to avoid
            # accidental printing
            _ = api.get_repo(self.name)
        except httpx.HTTPStatusError:
            return False
        else:
            return True

    def info(self, api: API) -> Dict[str, Any]:
        """
        Returns summary information about the repo
        from the API or Path.stat.

        Prefers the API info as it is more detailed.
        Will fall back to Path.stat only if the project
        is not available remotely.

        Args:
            api (API): GitHub API object.

        Returns:

            Dict[str, Union[str, int]]: Summary info.
        """

        info_dict: Dict[str, Any] = {}

        # Path.stat returns a UNIX timestamp for dates/times
        str_time_format: str = r"%Y-%m-%d %H:%M:%S"

        if self.exists_remote(api=api):
            # If remote, the API does all the work
            info_dict.update(api.get_repo_info(repo=self.name))
            info_dict["remote"] = True
            info_dict["local"] = self.exists_local()
        elif self.exists_local():
            # If local, we can get a bit of info from os.stat
            info_dict.update(
                {
                    "name": self.local_path.name,
                    "created_at": datetime.utcfromtimestamp(
                        self.local_path.stat().st_ctime
                    ).strftime(str_time_format),
                    "updated_at": datetime.utcfromtimestamp(
                        self.local_path.stat().st_mtime
                    ).strftime(str_time_format),
                    "size": self.local_path.stat().st_size,
                    "local": True,
                    "remote": self.exists_remote(api=api),
                },
            )
        else:
            raise RepoNotFoundError(
                f"Repo: {self.name!r} does not exist locally or remotely."
            )

        return info_dict

    def file_exists(self, file: str) -> bool:
        """
        Helper method to determine whether or not
        the repo root directory contains `file`.

        Args:
            file (str): Name of the file, relative
                to `self.local_path`.

        Returns:
            bool: True if exists, else False.
        """
        return self.local_path.joinpath(file).exists()

    def is_setuptools(self) -> bool:
        """
        Is the project based on setuptools,
        i.e. does it have a 'setup.cfg' or a 'setup.py'

        Returns:
            bool: True if setuptools, else False.
        """

        return self.file_exists("setup.cfg") or self.file_exists("setup.py")

    def has_pyproject_toml(self) -> bool:
        """
        Does the project have a 'pyproject.toml' file.

        Returns:
            bool: True if yes, else False.
        """

        return self.file_exists("pyproject.toml")

    def is_conda(self) -> bool:
        """
        Is the project based on conda,
        the only way really of detecting this automatically
        is to look for an 'environment.yml'

        Returns:
            bool: True if conda, else False.
        """
        return self.file_exists("environment.yml")

    def is_poetry(self) -> bool:
        """
        Does the project specify a poetry build_system.

        Returns:
            bool: True if yes, else False
        """

        # First check if it even has a pyproject.toml
        if not self.has_pyproject_toml():
            return False

        toml_dict = toml.load(self.local_path.joinpath("pyproject.toml"))
        build_system = toml_dict.get("build-system")
        # No build system means no poetry
        if not build_system:
            return False

        build_backend: str = build_system.get("build-backend")
        # No build backend means a bad toml file
        # also no poetry
        if not build_backend:
            return False

        # Now we know the toml file is valid
        # check if it specifies poetry
        return "poetry" in build_backend.strip().lower()

    def is_flit(self) -> bool:
        """
        Does the project specify a flit build_system.

        Returns:
            bool: True if yes, else False
        """

        # First check if it even has a pyproject.toml
        if not self.has_pyproject_toml():
            return False

        toml_dict = toml.load(self.local_path.joinpath("pyproject.toml"))
        build_system = toml_dict.get("build-system")
        # No build system, no flit
        if not build_system:
            return False

        build_backend: str = build_system.get("build-backend")
        # No build backend means bad toml file
        # also no flit
        if not build_backend:
            return False

        # Now we know the toml file is valid
        # check if it specifies flit
        return "flit" in build_backend.strip().lower()

    def dispatch_env(self) -> Optional[Environment]:
        """
        Returns the correct environment
        object for the calling `Repo` instance. Or `None` if it cannot
        detect the environment.

        Therefore all usage should first check for `None`.

        Returns:
            Optional[BaseEnvironment]: Either the correct environment
                object if it was able to detect. Or None.
        """

        if self.is_conda():
            return Conda(name=self.name, project_path=self.local_path)
        elif self.is_setuptools():
            return Venv(project_path=self.local_path)
        else:
            return None



"""
The pytoil gh command.

Author: Tom Fleet
Created: 25/06/2021
"""

import httpx
import typer
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config
from pytoil.repo import Repo

app = typer.Typer()


@app.command()
def gh(
    project: str = typer.Argument(..., help="Name of the project to go to on GitHub.")
) -> None:  # pragma: no cover
    """
    Open one of your projects on GitHub.

    Given a project name (must exist on GitHub and be owned by you),
    'gh' will open your browser and navigate to the project on GitHub.

    Examples:

    $ pytoil gh my_project
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    api = API(username=config.username, token=config.token)

    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )

    try:
        exists_remote = repo.exists_remote(api=api)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)
    else:

        if not exists_remote:
            msg.warn(
                f"{project!r} not found on GitHub! Was it a typo?", spaced=True, exits=1
            )
        else:
            msg.info(f"Opening {project!r} on GitHub...", spaced=True)
            typer.launch(url=repo.html_url)



"""
The pytoil docs command.

Author: Tom Fleet
Created: 25/06/2021
"""

import typer
from wasabi import msg

app = typer.Typer()


PYTOIL_DOCS_URL: str = "https://followtheprocess.github.io/pytoil/"


@app.command()
def docs() -> None:  # pragma: no cover
    """
    Open pytoil's documentation in your browser.

    Examples:

    $ pytoil docs
    """
    msg.info("Opening pytoil's docs in your browser...", spaced=True)
    typer.launch(url=PYTOIL_DOCS_URL)



"""
Abstract base class for all starter templates.

These templates are not supposed to be exhaustive, for that
the user is better off using pytoil's cookiecutter functionality.

The templates defined in the `starters` module are exactly that,
a starter. A good reference would be the behaviour of `cargo new`
in rust, which simply set's up a few basic sub directories
and a "hello world" function in main.rs.

Author: Tom Fleet
Created: 23/06/2021
"""


from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional


class BaseStarter(ABC):
    """
    Abstract base class defining the interface that all
    starter templates musti implement.
    """

    @property
    def path(self) -> Path:
        """
        `path` is the path on the filesystem under which to
        generate the template.

        e.g. config.projects_dir
        """
        raise NotImplementedError

    @property
    def name(self) -> str:
        """
        `name` is the name of the project, i.e. the name of the
        directory under `path` in which all the template's files
        and subdirectories should be created.

        e.g. projects_dir/name
        """
        raise NotImplementedError

    @property
    def root(self) -> Path:
        """
        The joined pathlib.Path of the project root.
        """

    @property
    def files(self) -> List[Path]:
        """
        Returns list of joined filepaths
        to be created.

        e.g. self.root.joinpath(filename)
        """
        raise NotImplementedError

    @abstractmethod
    def generate(self, username: Optional[str] = None) -> None:
        """
        Implements the generation of the project starter template.

        How it does this is up to the child class, it may directly
        instantiate files and directories using python, or call
        to external tools such as cargo, npm or go.
        """
        raise NotImplementedError



"""
The pytoil checkout command.

Author: Tom Fleet
Created: 25/06/2021
"""

import httpx
import typer
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config
from pytoil.exceptions import EnvironmentAlreadyExistsError
from pytoil.git import Git
from pytoil.repo import Repo
from pytoil.vscode import VSCode

app = typer.Typer()


@app.command()
def checkout(
    project: str = typer.Argument(
        ...,
        help="Name of the project to checkout.",
    ),
    venv: bool = typer.Option(
        False,
        "--venv",
        "-v",
        help="Attempt to auto-create a virtual environment.",
        show_default=False,
    ),
) -> None:
    """
    Checkout an existing development project.

    The checkout command lets you easily resume work on an existing project,
    whether that project is available locally in your configured projects
    directory, or if it is on GitHub.

    If pytoil finds your project locally, and you have enabled VSCode in your
    config file it will open it for you. If not, it will just tell you it already
    exists locally and where to find it.

    If your project is on your GitHub, pytoil will clone it for you and then open it
    (or tell you where it cloned it if you dont have VSCode set up).

    Finally, if checkout can't find a match after searching locally and on GitHub,
    it will prompt you to use 'pytoil new' to create a new one.

    If you pass the shorthand to someone elses repo e.g. 'someoneelse/repo' pytoil
    will detect this and automatically create a fork of this repo for you. Forking
    happens asynchronously so we can't clone your fork straight away. Give it a
    few seconds then a 'pytoil checkout repo' will bring it down as normal.

    You can also ask pytoil to automatically create a virtual environment on
    checkout with the '--venv/-v' flag. This only happens for projects pulled down
    from GitHub to avoid accidentally screwing up local projects.

    If the '--venv/-v' flag is used, pytoil will look at your project to try and detect
    which type of environment to create (conda or standard python).

    More info can be found in the documentation.

    Examples:

    $ pytoil checkout my_project

    $ pytoil checkout my_project --venv

    $ pytoil checkout someoneelse/project
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    # Setup the objects required
    api = API(username=config.username, token=config.token)
    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )
    code = VSCode(root=repo.local_path)
    git = Git()

    # Naively detect a "user/repo" type pattern
    # in which case fork the repo
    if "/" in project:
        owner, name = project.split("/")
        fork_repo(owner=owner, name=name, api=api, config=config)

    if repo.exists_local():
        # No environment or git stuff here, chances are if it exists locally
        # user has already done all this stuff
        checkout_local(repo=repo, code=code, config=config)

    elif repo.exists_remote(api=api):
        msg.info(f"{repo.name!r} found on GitHub. Cloning...", spaced=True)
        git.clone(url=repo.clone_url, check=True, cwd=config.projects_dir)

        env = repo.dispatch_env()

        if venv:
            if not env:
                msg.warn(
                    "Unable to auto-detect required environent. Skipping.",
                    spaced=True,
                )
            else:
                msg.info("Auto creating correct virtual environment.", spaced=True)

                try:
                    with msg.loading("Creating Environment..."):
                        env.create(packages=config.common_packages)
                except EnvironmentAlreadyExistsError:
                    msg.warn(
                        title="Environment already exists!",
                        text="No need to create a new one. Skipping.",
                    )
                finally:
                    if config.vscode:
                        code.set_workspace_python(python_path=env.executable)

        if config.vscode:
            msg.info(f"Opening {repo.name!r} in VSCode.", spaced=True)
            code.open()

    else:
        msg.warn(
            title=f"{repo.name!r} not found locally or on GitHub!",
            text=f"Does it exist? If not, create a new project with 'pytoil new {repo.name}'.",  # noqa: E501
            exits=1,
        )


def fork_repo(owner: str, name: str, api: API, config: Config) -> None:
    """
    Forks the passed repo and informs the user.

    Args:
        owner (str): Owner of the repo to be forked.
        name (str): Name of the repo to be forked.
        api (API): Configured API object.
        config (Config): Configured Config object.
    """

    if owner == config.username:
        msg.warn(
            "You don't need the '/' when checking out a repo you own!",
            spaced=True,
            exits=1,
        )

    msg.info(f"'{owner}/{name}' belongs to {owner!r}. Making you a fork.")
    try:
        api.create_fork(owner=owner, repo=name)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(err)

    # Forking happens asynchronously so we can't clone straight away
    # so just report success and exit
    msg.good(
        "Done!",
        text="Note: Forking happens asynchronously on GitHub which means "
        "it may not be available to clone right away.",
        spaced=True,
        exits=0,
    )


def checkout_local(repo: Repo, code: VSCode, config: Config) -> None:
    """
    Helper function to checkout a local repo.

    Args:
        repo (Repo): Repo to checkout.
        code (VSCode): VSCode instance to handle opening.
        config (Config): Configured Config object.
    """

    msg.info(f"{repo.name!r} available locally.", spaced=True)

    if config.vscode:
        msg.text(f"Opening {repo.name!r} in VSCode.")
        code.open()
